"""The components of the derivative of the entropy
for the Choo and Siow homoskedastic model.
"""
from typing import Optional, Tuple, Union
import numpy as np

from .utils import bs_error_abort
from .matching_utils import Matching
from .entropy import EntropyFunctions, EntropyHessianComponents


def _entropy_choo_siow(
    muhat: Matching, deriv: Optional[int] = 0
) -> Union[
    np.ndarray,
    Tuple[np.ndarray, np.ndarray],
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
]:
    """Returns the values of $\mathcal{E}$
    and the first (if `deriv` is 1 or 2) and second (if `deriv` is 2) derivatives
    for the Choo and Siow model

    Args:
        muhat: a Matching
        deriv: if equal 1, we compute the first derivatives too;
               if equals 2, also the hessian

    Returns:
        the value of the generalized entropy
        if deriv = 1 or 2, the (X,Y) matrix of the first derivative of the entropy
        if deriv = 2, the (X,Y,X,Y) array of the second derivative
            wrt $(\mu,\mu)$
          and the (X,Y,X+Y) second derivatives
            wrt $(\mu,(n,m))$
    """
    muxy, mux0, mu0y, n, m = muhat.unpack()

    logxy = np.log(muxy)
    logx0 = np.log(mux0)
    log0y = np.log(mu0y)

    val_entropy = (
        -2.0 * np.sum(muxy * logxy)
        - np.sum(mux0 * logx0)
        - np.sum(mu0y * log0y)
        + np.sum(n * np.log(n))
        + np.sum(m * np.log(m))
    )

    if deriv == 0:
        return val_entropy
    if deriv in [1, 2]:
        der_xy = -2.0 * logxy + log0y
        der_xy += logx0.reshape((-1, 1))
        if deriv == 1:
            return val_entropy, der_xy
        else:  # we compute the Hessians
            X, Y = muxy.shape
            derlogxy = 1.0 / muxy
            derlogx0 = 1.0 / mux0
            derlog0y = 1.0 / mu0y
            der2_xyzt = np.zeros((X, Y, X, Y))
            der2_xyr = np.zeros((X, Y, X + Y))
            for x in range(X):
                dlogx0 = derlogx0[x]
                for y in range(Y):
                    d2xy = np.zeros((X, Y))
                    d2xy[x, :] = -dlogx0
                    d2xy[:, y] -= derlog0y[y]
                    d2xy[x, y] -= 2.0 * derlogxy[x, y]
                    der2_xyzt[x, y, :, :] = d2xy
                    der2_xyr[x, y, x] = derlogx0[x]
                    der2_xyr[x, y, X + y] = derlog0y[y]
            return val_entropy, der_xy, der2_xyzt, der2_xyr
    else:
        bs_error_abort("deriv should be 0, 1, or 2")


def e0_fun_choo_siow(muhat: Matching) -> np.ndarray:
    """Returns the values of $e_0$ for the Choo and Siow model.

    Args:
        muhat: a Matching

    Returns:
        the (X,Y) matrix of the first derivative of the entropy
    """
    entropy_res = _entropy_choo_siow(muhat, deriv=1)
    return entropy_res[1]


def e0_derivative_choo_siow(muhat: Matching) -> EntropyHessianComponents:
    """Returns the derivatives of $e_0$ for the Choo and Siow model.

    Args:
        muhat: a Matching

    Returns:
        the three components of the hessian wrt $(\mu,\mu)$ of the entropy
        and the two components of the hessian wrt $(\mu,r)$
    """
    entropy_res = _entropy_choo_siow(muhat, deriv=2)
    hessmumu = entropy_res[2]
    hessmur = entropy_res[3]
    muxy, *_ = muhat.unpack()
    X, Y = muxy.shape
    hess_x = np.zeros((X, Y, Y))
    hess_y = np.zeros((X, Y, X))
    hess_xy = np.zeros((X, Y))
    hess_nx = np.zeros((X, Y))
    hess_my = np.zeros((X, Y))
    for x in range(X):
        for y in range(Y):
            d2xy = hessmumu[x, y, :, :]
            d2r = hessmur[x, y, :]
            hess_x[x, y, :] = d2xy[x, :]
            hess_y[x, y, :] = d2xy[:, y]
            hess_xy[x, y] = d2xy[x, y]
            hess_nx[x, y] = d2r[x]
            hess_my[x, y] = d2r[X + y]
    return (hess_x, hess_y, hess_xy), (hess_nx, hess_my)


entropy_choo_siow = EntropyFunctions(
    e0_fun=e0_fun_choo_siow,
    hessian="provided",
    e0_derivative=e0_derivative_choo_siow,
    description="Choo and Siow homoskedastic with analytic Hessian",
)

entropy_choo_siow_numeric = EntropyFunctions(
    e0_fun=e0_fun_choo_siow,
    description="Choo and Siow homoskedastic with numerical Hessian",
)



"""Estimates the semilinear Choo and Siow homoskedastic (2006) model
using Poisson GLM.
"""

from typing import Optional
from math import sqrt
import numpy as np
import scipy.linalg as spla
import scipy.sparse as spr
from sklearn import linear_model

from .matching_utils import Matching, _make_XY_K_mat, _variance_muhat
from .poisson_glm_utils import PoissonGLMResults, _prepare_data


def choo_siow_poisson_glm(
    muhat: Matching,
    phi_bases: np.ndarray,
    tol: Optional[float] = 1e-12,
    max_iter: Optional[int] = 10000,
    verbose: Optional[int] = 1,
) -> PoissonGLMResults:
    """Estimates the semilinear Choo and Siow homoskedastic (2006) model
        using Poisson GLM.

    Args:
        muhat: the observed Matching
        phi_bases: an (X, Y, K) array of bases
        tol: tolerance level for `linear_model.PoissonRegressor.fit`
        max_iter: maximum number of iterations
            for `linear_model.PoissonRegressor.fit`
        verbose: defines how much output we want (0 = least)

    Returns:
        a `PoissonGLMResults` instance

    Example:
        ```py
        n_households = 1e6
        X, Y, K = 4, 3, 6
        # we setup a quadratic set of basis functions
        phi_bases = np.zeros((X, Y, K))
        phi_bases[:, :, 0] = 1
        for x in range(X):
            phi_bases[x, :, 1] = x
            phi_bases[x, :, 3] = x * x
            for y in range(Y):
                phi_bases[x, y, 4] = x * y
        for y in range(Y):
            phi_bases[:, y, 2] = y
            phi_bases[:, y, 5] = y * y

        lambda_true = np.random.randn(K)
        phi_bases = np.random.randn(X, Y, K)
        Phi = phi_bases @ lambda_true

        # we simulate a Choo and Siow sample from a population
        #  with equal numbers of men and women of each type
        n = np.ones(X)
        m = np.ones(Y)
        choo_siow_instance = ChooSiowPrimitives(Phi, n, m)
        mus_sim = choo_siow_instance.simulate(n_households)
        muxy_sim, mux0_sim, mu0y_sim, n_sim, m_sim = mus_sim.unpack()

        results = choo_siow_poisson_glm(mus_sim, phi_bases)

        # compare true and estimated parameters
        results.print_results(
            lambda_true,
            u_true=-np.log(mux0_sim / n_sim),
            v_true=-np.log(mu0y_sim / m_sim)
        )
        ```

    """
    try_sparse = False
    X, Y, K = phi_bases.shape
    XY = X * Y
    n_rows = XY + X + Y
    n_cols = X + Y + K

    # the vector of weights for the Poisson regression
    w = np.concatenate((2 * np.ones(XY), np.ones(X + Y)))
    # reshape the bases
    phi_mat = _make_XY_K_mat(phi_bases)

    if try_sparse:
        w_mat = spr.csr_matrix(
            np.concatenate((2 * np.ones(XY, n_cols), np.ones(X + Y, n_cols)))
        )

        # construct the Z matrix
        ones_X = spr.csr_matrix(np.ones((X, 1)))
        ones_Y = spr.csr_matrix(np.ones((Y, 1)))
        zeros_XK = spr.csr_matrix(np.zeros((X, K)))
        zeros_YK = spr.csr_matrix(np.zeros((Y, K)))
        zeros_XY = spr.csr_matrix(np.zeros((X, Y)))
        zeros_YX = spr.csr_matrix(np.zeros((Y, X)))
        id_X = spr.csr_matrix(np.eye(X))
        id_Y = spr.csr_matrix(np.eye(Y))
        Z_unweighted = spr.vstack(
            [
                spr.hstack([-spr.kron(id_X, ones_Y), -spr.kron(ones_X, id_Y), phi_mat]),
                spr.hstack([-id_X, zeros_XY, zeros_XK]),
                spr.hstack([zeros_YX, -id_Y, zeros_YK]),
            ]
        )
        Z = Z_unweighted / w_mat
    else:
        ones_X = np.ones((X, 1))
        ones_Y = np.ones((Y, 1))
        zeros_XK = np.zeros((X, K))
        zeros_YK = np.zeros((Y, K))
        zeros_XY = np.zeros((X, Y))
        zeros_YX = np.zeros((Y, X))
        id_X = np.eye(X)
        id_Y = np.eye(Y)
        Z_unweighted = np.vstack(
            [
                np.hstack([-np.kron(id_X, ones_Y), -np.kron(ones_X, id_Y), phi_mat]),
                np.hstack([-id_X, zeros_XY, zeros_XK]),
                np.hstack([zeros_YX, -id_Y, zeros_YK]),
            ]
        )
        Z = Z_unweighted / w.reshape((-1, 1))

    _, _, _, n, m = muhat.unpack()
    var_muhat, var_munm = _variance_muhat(muhat)
    (
        muxyhat_norm,
        var_muhat_norm,
        var_munm_norm,
        n_households,
        n_individuals,
    ) = _prepare_data(muhat, var_muhat, var_munm)

    clf = linear_model.PoissonRegressor(
        fit_intercept=False, tol=tol, verbose=verbose, alpha=0, max_iter=max_iter
    )
    clf.fit(Z, muxyhat_norm, sample_weight=w)
    gamma_est = clf.coef_

    # we compute the variance-covariance of the estimator
    nr, nc = Z.shape
    exp_Zg = np.exp(Z @ gamma_est).reshape(n_rows)
    A_hat = np.zeros((nc, nc))
    B_hat = np.zeros((nc, nc))
    for i in range(nr):
        Zi = Z[i, :]
        wi = w[i]
        A_hat += wi * exp_Zg[i] * np.outer(Zi, Zi)
        for j in range(nr):
            Zj = Z[j, :]
            B_hat += wi * w[j] * var_muhat_norm[i, j] * np.outer(Zi, Zj)

    A_inv = spla.inv(A_hat)
    variance_gamma = A_inv @ B_hat @ A_inv
    stderrs_gamma = np.sqrt(np.diag(variance_gamma))

    beta_est = gamma_est[-K:]
    beta_std = stderrs_gamma[-K:]
    Phi_est = phi_bases @ beta_est

    # we correct for the effect of the normalization
    n_norm = n / n_individuals
    m_norm = m / n_individuals
    u_est = gamma_est[:X] + np.log(n_norm)
    v_est = gamma_est[X:-K] + np.log(m_norm)

    # since u = a + log(n_norm) we also need to adjust the estimated variance
    z_unweighted_T = Z_unweighted.T
    u_std = np.zeros(X)
    ix = XY
    for x in range(X):
        n_norm_x = n_norm[x]
        A_inv_x = A_inv[x, :]
        var_log_nx = var_munm_norm[ix, ix] / n_norm_x / n_norm_x
        slice_x = slice(x * Y, (x + 1) * Y)
        covar_term = var_muhat_norm[:, ix] + np.sum(var_muhat_norm[:, slice_x], 1)
        cov_a_lognx = (A_inv_x @ z_unweighted_T @ covar_term) / n_norm_x
        ux_var = variance_gamma[x, x] + var_log_nx + 2.0 * cov_a_lognx
        u_std[x] = sqrt(ux_var)
        ix += 1

    v_std = stderrs_gamma[X:-K]
    iy, jy = X, XY + X
    for y in range(Y):
        m_norm_y = m_norm[y]
        A_inv_y = A_inv[iy, :]
        var_log_my = var_munm_norm[jy, jy] / m_norm_y / m_norm_y
        slice_y = slice(y, XY, Y)
        covar_term = var_muhat_norm[:, jy] + np.sum(var_muhat_norm[:, slice_y], 1)
        cov_b_logmy = (A_inv_y @ z_unweighted_T @ covar_term) / m_norm_y
        vy_var = variance_gamma[iy, iy] + var_log_my + 2.0 * cov_b_logmy
        v_std[y] = sqrt(vy_var)
        iy += 1
        jy += 1

    results = PoissonGLMResults(
        X=X,
        Y=Y,
        K=K,
        number_households=n_households,
        number_individuals=n_individuals,
        estimated_gamma=gamma_est,
        estimated_Phi=Phi_est,
        estimated_beta=beta_est,
        estimated_u=u_est,
        estimated_v=v_est,
        variance_gamma=variance_gamma,
        stderrs_gamma=stderrs_gamma,
        stderrs_beta=beta_std,
        stderrs_u=u_std,
        stderrs_v=v_std,
    )

    return results



"""The components of the derivative of the entropy
for the Choo and Siow gender-heteroskedastic model.

We normalize the standard error for the X side at 1,
and we estimate the standard error on the Y side.
"""
import numpy as np

from .entropy import EntropyFunctions, EntropyHessianComponents
from .matching_utils import Matching


def e0_choo_siow_gender_heteroskedastic(muhat: Matching) -> np.ndarray:
    """Returns the values of the parameter-independent part $e_0$
    for the Choo and Siow gender-heteroskedastic model; we normalized $\sigma=1$.

    Args:
        muhat: a Matching

    Returns:
        the (X,Y) matrix of the parameter-independent part
        of the first derivative of the entropy.
    """
    muxy, mux0, *_ = muhat.unpack()
    e0_vals = -np.log(muxy / mux0.reshape((-1, 1)))
    return e0_vals


def e0_derivative_choo_siow_gender_heteroskedastic(
    muhat: Matching,
) -> EntropyHessianComponents:
    """Returns the derivatives of the parameter-independent part $e_0$
    for the Choo and Siow gender-heteroskedastic model; we normalized $\sigma=1$.

    Args:
        muhat: a Matching

    Returns:
        the  components of the parameter-independent part of the hessian of the entropy.
    """
    muxy, mux0, *_ = muhat.unpack()
    X, Y = muxy.shape
    hess_x = np.zeros((X, Y, Y))
    hess_y = np.zeros((X, Y, X))
    hess_xy = np.zeros((X, Y))
    hess_n = np.zeros((X, Y))
    hess_m = np.zeros((X, Y))
    der_logxy = 1.0 / muxy
    der_logx0 = 1.0 / mux0
    for x in range(X):
        dlogx0 = der_logx0[x]
        for y in range(Y):
            hess_x[x, y, :] = -dlogx0
            hess_xy[x, y] = -der_logxy[x, y] - dlogx0
            hess_n[x, y] = dlogx0
    return (hess_x, hess_y, hess_xy), (hess_n, hess_m)


def e_choo_siow_gender_heteroskedastic(muhat: Matching) -> np.ndarray:
    """Returns the values of the parameter-dependent part  $e$
    for the Choo and Siow gender-heteroskedastic model; we normalized $\sigma=1$.

    Args:
        muhat: a Matching

    Returns:
        the (X,Y,1) array of the parameter-dependent part
        of the first derivative of the entropy.
    """
    muxy, _, mu0y, *_ = muhat.unpack()
    X, Y = muxy.shape
    n_alpha = 1

    e_vals = np.zeros((X, Y, n_alpha))
    e_vals[:, :, 0] = -np.log(muxy / mu0y)
    return e_vals


def e_derivative_choo_siow_gender_heteroskedastic(
    muhat: Matching,
) -> EntropyHessianComponents:
    """Returns the derivatives of the parameter-dependent part $e$
     for the Choo and Siow gender-heteroskedastic model;
     we normalized $\sigma_1=1$.

    Args:
        muhat: a Matching

    Returns:
        the components of the parameter-dependent part of the hessian of the entropy.
    """
    muxy, _, mu0y, *_ = muhat.unpack()
    X, Y = muxy.shape

    n_alpha = 1
    hess_x = np.zeros((X, Y, Y, n_alpha))
    hess_y = np.zeros((X, Y, X, n_alpha))
    hess_xy = np.zeros((X, Y, n_alpha))
    hess_n = np.zeros((X, Y, n_alpha))
    hess_m = np.zeros((X, Y, n_alpha))
    der_logxy = 1.0 / muxy
    der_log0y = 1.0 / mu0y
    for x in range(X):
        for y in range(Y):
            dlog0y = der_log0y[y]
            hess_y[x, y, :, 0] = -dlog0y
            hess_xy[x, y, 0] = -der_logxy[x, y] - dlog0y
            hess_m[x, y, 0] = dlog0y
    return (hess_x, hess_y, hess_xy), (hess_n, hess_m)


entropy_choo_siow_gender_heteroskedastic = EntropyFunctions(
    e0_fun=e0_choo_siow_gender_heteroskedastic,
    parameter_dependent=True,
    e_fun=e_choo_siow_gender_heteroskedastic,
    hessian="provided",
    e0_derivative=e0_derivative_choo_siow_gender_heteroskedastic,
    e_derivative=e_derivative_choo_siow_gender_heteroskedastic,
    description="Choo and Siow gender-heteroskedastic with analytic Hessian",
)

entropy_choo_siow_gender_heteroskedastic_numeric = EntropyFunctions(
    e0_fun=e0_choo_siow_gender_heteroskedastic,
    parameter_dependent=True,
    e_fun=e_choo_siow_gender_heteroskedastic,
    description="Choo and Siow gender-heteroskedastic with numerical Hessian",
)



"""Entropies of some useful models. """
from dataclasses import dataclass
from functools import partial
from typing import Callable, List, Optional, Tuple
import numpy as np

from .utils import _EPS, _TWO_EPS, bs_error_abort
from .matching_utils import Matching


EntropyGradient = Callable[[Matching, Optional[List]], np.ndarray]
"""The type of a function that takes in a Matching
   and possibly a list of other parameters
   and returns an array with the gradient of entropy wrt $\mu$.
"""

EntropyHessianComponentsMuMu = Tuple[np.ndarray, np.ndarray, np.ndarray]
"""The type of a tuple of the three components of the hessian of the entropy
   wrt $(\mu,\mu)$.
"""

EntropyHessianMuMu = Callable[[Matching], EntropyHessianComponentsMuMu]
"""The type of a function that takes in a Matching
   and returns the three components of the hessian of the entropy
   wrt $(\mu,\mu)$.
"""


EntropyHessianComponentsMuR = Tuple[np.ndarray, np.ndarray]
"""The type of a tuple of the two components of the hessian of the entropy
   wrt $(\mu,n)$ and $(\mu, m))$.
"""

EntropyHessianMuR = Callable[[Matching], EntropyHessianComponentsMuR]
"""The type of a function that takes in a Matching
   and returns the two components of the hessian of the entropy
   wrt $(\mu,n)$ and $(\mu, m))$.
"""

EntropyHessianComponents = Tuple[
    EntropyHessianComponentsMuMu, EntropyHessianComponentsMuR
]
""" combines the tuples of components of the hessians"""

EntropyHessians = Tuple[EntropyHessianMuMu, EntropyHessianMuR]
""" combines the hessians """


@dataclass
class EntropyFunctions:
    """Defines the entropy used, via the derivative $e_0 + e \cdot \\alpha$

    Attributes:
        e0_fun: required
        parameter_dependent:  if `True`, the entropy depends on parameters.
            Defaults to `False`
        e_fun: only in entropies that depend on parameters.
            Defaults to `None`
        hessian: defaults to `"numeric"`
            * if `"provide"`, we provide the hessian of the entropy.
            * if `"numerical"`, it is computed by central differences.
        e0_derivative: the derivative of `e0_fun`, if available.
            Defaults to `None`
        e_derivative: the derivative of `e_fun`, if available.
            Defaults to `None`
        more_params: additional parameters
            that define the distribution of errors.
            Defaults to `None`
        description: some text describing the model.
            Defaults to `None`

    Examples:
        See `entropy_choo_siow` in `choo_siow.py`
    """

    e0_fun: EntropyGradient
    parameter_dependent: Optional[bool] = False
    e_fun: Optional[EntropyGradient] = None
    hessian: Optional[str] = "numerical"
    e0_derivative: Optional[EntropyHessians] = None
    e_derivative: Optional[EntropyHessians] = None
    more_params: Optional[List] = None
    description: Optional[str] = None

    def __post_init__(self):
        if not self.parameter_dependent:
            if self.hessian == "provided" and self.e0_derivative is None:
                bs_error_abort(
                    "You claim to provide the hessian "
                    + "but you did not provide the e0_derivative."
                )
        else:
            if self.e_fun is None:
                bs_error_abort(
                    "Your entropy is parameter dependent "
                    + " but you did not provide the e_fun."
                )
            if self.hessian == "provided" and self.e_derivative is None:
                bs_error_abort(
                    "Your entropy is parameter dependent, "
                    + "you claim to provide the hessian,\n"
                    + " but I do not see the e_derivative."
                )


def entropy_gradient(
    entropy: EntropyFunctions,
    muhat: Matching,
    alpha: Optional[np.ndarray] = None,
    more_params: Optional[List] = None,
) -> np.ndarray:
    """Computes the derivative of the entropy wrt $\mu$
     at $(\mu, n, m, \alpha)$

    Args:
        entropy: the `EntropyFunctions` object
        muhat: a Matching
        alpha: a vector of parameters of the derivative of the entropy, if any
        more_params: a list of additional parameters, if any

    Returns:
        the derivative of the entropy wrt $\mu$ at $(\mu, n, m, \alpha)$.
    """
    e0_fun = entropy.e0_fun
    if more_params is None:
        e0_vals = e0_fun(muhat)
    else:
        e0_vals = e0_fun(muhat, more_params=more_params)
    parameter_dependent = entropy.parameter_dependent
    if parameter_dependent:
        if alpha is None:
            bs_error_abort("alpha should be specified for this model")
        e_fun = entropy.e_fun
        if more_params is None:
            e_vals = e_fun(muhat)
        else:
            e_vals = e_fun(muhat, more_params=more_params)
        return e0_vals + e_vals @ alpha
    else:
        return e0_vals


def _numeric_hessian(
    entropy: EntropyFunctions,
    muhat: Matching,
    alpha: Optional[np.ndarray] = None,
    more_params: Optional[List] = None,
) -> EntropyHessianComponents:
    """Evaluates numerically the components of the hessians of the entropy
    wrt $(\mu,\mu)$ and $(\mu,(n,m))$

    Args:
        entropy: the `EntropyFunctions` object
        muhat: a Matching
        alpha: a vector of parameters of the derivative of the entropy, if any
        more_params: a list of additional parameters, if any

    Returns:
        the hessians of the entropy wrt $(\mu,\mu)$ and $(\mu,(n,m))$.
    """
    parameter_dependent = entropy.parameter_dependent
    if not parameter_dependent:
        if more_params is None:
            entropy_deriv = entropy.e0_fun
        else:
            entropy_deriv = partial(entropy_gradient, entropy, more_params=more_params)
    else:
        if more_params is None:
            entropy_deriv = partial(entropy_gradient, entropy, alpha=alpha)
        else:
            entropy_deriv = partial(
                entropy_gradient, entropy, alpha=alpha, more_params=more_params
            )

    muxyhat, _, _, n, m = muhat.unpack()
    X, Y = muxyhat.shape

    # start with the hessian wrt (mu, mu)
    hessian_x = np.zeros((X, Y, Y))
    hessian_y = np.zeros((X, Y, X))
    hessian_xy = np.zeros((X, Y))
    for x in range(X):
        for y in range(Y):
            for t in range(Y):
                muxy = muxyhat.copy().astype(float)
                muxy[x, t] += _EPS
                mus = Matching(muxy, n, m)
                der_entropy_plus = entropy_deriv(mus)
                muxy[x, t] -= _TWO_EPS
                mus = Matching(muxy, n, m)
                der_entropy_minus = entropy_deriv(mus)
                hessian_x[x, y, t] = (
                    der_entropy_plus[x, y] - der_entropy_minus[x, y]
                ) / _TWO_EPS
            for z in range(X):
                muxy = muxyhat.copy().astype(float)
                muxy[z, y] += _EPS
                mus = Matching(muxy, n, m)
                der_entropy_plus = entropy_deriv(mus)
                muxy[z, y] -= _TWO_EPS
                mus = Matching(muxy, n, m)
                der_entropy_minus = entropy_deriv(mus)
                hessian_y[x, y, z] = (
                    der_entropy_plus[x, y] - der_entropy_minus[x, y]
                ) / _TWO_EPS

            muxy = muxyhat.copy().astype(float)
            muxy[x, y] += _EPS
            mus = Matching(muxy, n, m)
            der_entropy_plus = entropy_deriv(mus)
            muxy[x, y] -= _TWO_EPS
            mus = Matching(muxy, n, m)
            der_entropy_minus = entropy_deriv(mus)
            hessian_xy[x, y] = (
                der_entropy_plus[x, y] - der_entropy_minus[x, y]
            ) / _TWO_EPS

    components_mumu = (hessian_x, hessian_y, hessian_xy)

    # now the hessian wrt (mu, r)
    hessian_n = np.zeros((X, Y))
    hessian_m = np.zeros((X, Y))
    for x in range(X):
        for y in range(Y):
            n1 = n.copy().astype(float)
            n1[x] += _EPS
            mus = Matching(muxyhat, n1, m)
            der_entropy_plus = entropy_deriv(mus)
            n1[x] -= _TWO_EPS
            mus = Matching(muxy, n1, m)
            der_entropy_minus = entropy_deriv(mus)
            hessian_n[x, y] = (
                der_entropy_plus[x, y] - der_entropy_minus[x, y]
            ) / _TWO_EPS

            m1 = m.copy().astype(float)
            m1[y] += _EPS
            mus = Matching(muxyhat, n, m1)
            der_entropy_plus = entropy_deriv(mus)
            m1[y] -= _TWO_EPS
            mus = Matching(muxyhat, n, m1)
            der_entropy_minus = entropy_deriv(mus)
            hessian_m[x, y] = (
                der_entropy_plus[x, y] - der_entropy_minus[x, y]
            ) / _TWO_EPS

    components_mur = (hessian_n, hessian_m)

    return components_mumu, components_mur


def _fill_hessianMuMu_from_components(
    hessian_components: EntropyHessianComponentsMuMu,
) -> np.ndarray:
    """Fills the hessian of the entropy wrt $(\mu,\mu)$

    Args:
        hessian_components: the three components of the hessian

    Returns:
        the (XY,XY) matrix of the hessian
    """
    hess_x, hess_y, hess_xy = hessian_components
    X, Y = hess_xy.shape
    XY = X * Y
    hessian = np.zeros((XY, XY))

    i = 0
    ix = 0
    for x in range(X):
        for y in range(Y):
            hessian[i, ix : (ix + Y)] = hess_x[x, y, :]
            slice_y = slice(y, XY, Y)
            hessian[i, slice_y] = hess_y[x, y, :]
            hessian[i, i] = hess_xy[x, y]
            i += 1
        ix += Y

    return hessian


def _fill_hessianMuR_from_components(
    hessian_components: EntropyHessianComponentsMuR,
) -> np.ndarray:
    """Fills the hessian of the entropy wrt $(\mu,(n,m))$

    Args:
        hessian_components: the two components of the hessian

    Returns:
        the (XY,X+Y) matrix of the hessian
    """
    hess_nx, hess_my = hessian_components
    X, Y = hess_nx.shape[:2]
    XY = X * Y
    hessian = np.zeros((XY, X + Y))

    i = 0
    for x in range(X):
        iy = X
        for y in range(Y):
            hessian[i, x] = hess_nx[x, y]
            hessian[i, iy] = hess_my[x, y]
            i += 1
            iy += 1

    return hessian



""" matching-related utilities """
from dataclasses import dataclass, field
from typing import List, Tuple
import numpy as np

from .utils import bs_error_abort, print_stars, test_matrix, test_vector


def _get_singles(
    muxy: np.ndarray, n: np.ndarray, m: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Computes the numbers of singles from the matches and the margins."""
    mux0 = n - np.sum(muxy, 1)
    mu0y = m - np.sum(muxy, 0)
    return mux0, mu0y


def _compute_margins(
    muxy: np.ndarray, mux0: np.ndarray, mu0y: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Computes the margins from the matches and the singles."""
    n = np.sum(muxy, 1) + mux0
    m = np.sum(muxy, 0) + mu0y
    return n, m


@dataclass
class Matching:
    """stores the numbers of couples and singles of every type;

    muxy is an (X,Y)-matrix
    n is an X-vector
    m is an Y-vector
    mux0 and mu0y are generated as the corresponding numbers of singles
    """

    mux0: np.ndarray = field(init=False)
    mu0y: np.ndarray = field(init=False)
    muxy: np.ndarray
    n: np.ndarray
    m: np.ndarray

    def __str__(self):
        X, Y = self.muxy.shape
        n_couples = np.sum(self.muxy)
        n_men, n_women = np.sum(self.n), np.sum(self.m)
        repr_str = f"This is a matching with {n_men}  men, {n_women} single women.\n"
        repr_str += f"   with {n_couples} couples,\n \n"
        repr_str += f" We have {X} types of men and {Y} of women."
        print_stars(repr_str)

    def __post_init__(self):
        X, Y = test_matrix(self.muxy)
        Xn = test_vector(self.n)
        Ym = test_vector(self.m)
        if Xn != X:
            bs_error_abort(f"muxy is a ({X}, {Y}) matrix but n has {Xn} elements.")
        if Ym != Y:
            bs_error_abort(f"muxy is a ({X}, {Y}) matrix but m has {Ym} elements.")
        self.mux0, self.mu0y = _get_singles(self.muxy, self.n, self.m)

    def unpack(self):
        return self.muxy, self.mux0, self.mu0y, self.n, self.m


def _get_margins(mus: Matching) -> Tuple[np.ndarray, np.ndarray]:
    """computes the numbers of each type from the matching patterns"""
    _, _, _, n, m = mus.unpack()
    return n, m


def _simulate_sample_from_mus(
    mus: Matching, n_households: int, seed: int = None
) -> Matching:
    """Draw a sample of n_households form the matching patterns in mus

    Args:
        mus: the matching patterns
        n_households: the number of households requested
        seed: an integer seed for the random number generator

    Returns:
        the sample matching patterns
    """
    rng = np.random.default_rng(seed)
    muxy, mux0, mu0y, _, _ = mus.unpack()
    X, Y = muxy.shape
    # stack all probabilities
    XY = X * Y
    num_choices = XY + X + Y
    pvec = np.zeros(num_choices)
    pvec[:XY] = muxy.reshape(XY)
    pvec[XY : (XY + X)] = mux0
    pvec[(XY + X) :] = mu0y
    pvec /= np.sum(pvec)
    matches = rng.multinomial(n_households, pvec)
    muxy_sim = matches[:XY].reshape((X, Y))
    mux0_sim = matches[XY : (XY + X)]
    mu0y_sim = matches[(XY + X) :]
    # make sure we have no zeros
    _MU_EPS = min(1, int(1e-3 * n_households))
    muxy_sim += _MU_EPS
    mux0_sim += _MU_EPS
    mu0y_sim += _MU_EPS
    n_sim, m_sim = _compute_margins(muxy_sim, mux0_sim, mu0y_sim)
    mus_sim = Matching(muxy=muxy_sim, n=n_sim, m=m_sim)
    return mus_sim


def _variance_muhat(muhat: Matching) -> Tuple[np.ndarray, np.ndarray]:
    """
    Computes the unweighted variance-covariance matrix of the observed matching patterns

    Args:
        muhat: a Matching object

    Returns:
        an (XY+X+Y, XY+X+Y) matrix of the variance-covariance
        of $(\mu_{xy}, \mu_{x0}, \mu_{0y})$ (row-major order)
        and an (X\mu_{xy}, n_x, m_y)$ (row-major order)
    """
    muxy, mux0, mu0y, n, m = muhat.unpack()

    # normalize all proportions
    n_households = np.sum(n) + np.sum(m) - np.sum(muxy)
    muxy_norm = (muxy / n_households).ravel()
    mux0_norm = mux0 / n_households
    mu0y_norm = mu0y / n_households

    X, Y = muxy.shape
    XY = X * Y
    var_dim = XY + X + Y
    var_muhat = np.zeros((var_dim, var_dim))

    # we start by constructing the variance of (muxy, mux0, mu0y)

    # variance of muxy
    varxy_norm = np.diag(muxy_norm) - np.outer(muxy_norm, muxy_norm)
    var_muhat[:XY, :XY] = varxy_norm
    # covariance of muxy and mux0
    cov_xy_x0 = -np.outer(muxy_norm, mux0_norm)
    var_muhat[:XY, XY : (XY + X)] = cov_xy_x0
    var_muhat[XY : (XY + X), :XY] = cov_xy_x0.T
    # covariance of muxy and mu0y
    cov_xy_0y = -np.outer(muxy_norm, mu0y_norm)
    var_muhat[:XY, (XY + X) :] = cov_xy_0y
    var_muhat[(XY + X) :, :XY] = cov_xy_0y.T
    # variance of mux0
    varx0_norm = np.diag(mux0_norm) - np.outer(mux0_norm, mux0_norm)
    var_muhat[XY : (XY + X), XY : (XY + X)] = varx0_norm
    # covariance of mux0 and mu0y
    cov_x0_0y = -np.outer(mux0_norm, mu0y_norm)
    var_muhat[XY : (XY + X), (XY + X) :] = cov_x0_0y
    var_muhat[(XY + X) :, XY : (XY + X)] = cov_x0_0y.T
    # variance of mu0y
    var0y_norm = np.diag(mu0y_norm) - np.outer(mu0y_norm, mu0y_norm)
    var_muhat[(XY + X) :, (XY + X) :] = var0y_norm

    # change of variables from (muxy, mux0, mu0y)  to (muxy, n, m)
    jac = np.eye(var_dim)
    ix = 0
    for x in range(X):
        jac[XY + x, ix : (ix + Y)] = 1
        ix += Y
    iy = XY + X
    for y in range(Y):
        slice_y = slice(y, XY, Y)
        jac[iy, slice_y] = 1
        iy += 1
    var_munm = jac @ var_muhat @ jac.T

    return n_households * var_muhat, n_households * var_munm


def _make_XY_K_mat(xyk_array: np.ndarray) -> np.ndarray:
    """Reshapes an (X,Y,K) array to an (XY,K) matrix.

    Args:
        xyk_array: an (X, Y, K) array of bases

    Returns:
        the same,  (XY, K)-reshaped
    """
    X, Y, K = xyk_array.shape
    XY = X * Y
    xy_k_mat = np.zeros((XY, K))
    for k in range(K):
        xy_k_mat[:, k] = xyk_array[:, :, k].ravel()
    return xy_k_mat


def _reshape4_to2(array4: np.ndarray) -> np.ndarray:
    """Reshapes an array (X,Y,Z,T) to a matrix (XY,ZT).

    Args:
        array4: an (X, Y, Z, T) array

    Returns:
        the same,  (XY, ZT)-reshaped
    """
    if array4.ndim != 4:
        bs_error_abort(f"array4 should have 4 dimensions not {array4.ndim}")
    X, Y, Z, T = array4.shape
    XY, ZT = X * Y, Z * T
    array2 = np.zeros((XY, ZT))
    xy = 0
    for x in range(X):
        for y in range(Y):
            array2[xy, :] = array4[x, y, :, :].ravel()
            xy += 1
    return array2


def _change_indices(nests: List[List[int]]):
    """subtracts 1 from the indices within the nest structure

    Args:
        nests: the nest structure

    Returns:
        a similar list
    """
    return [[nest_i - 1 for nest_i in nest] for nest in nests]


def _find_nest_of(nests: List[List[int]], y: int) -> int:
    """find the index of the nest that contains y, or return -1

    Args:
        nests: a nest structure
        y: the type we are looking for

    Returns:
        the nest of y, or -1 if not found
    """
    for i_n, nest in enumerate(nests):
        if y in nest:
            return i_n
    return -1  # if not found



"""
To test the quality of the estimators, we generate data
both from a semilinear Choo and Siow  model
and from a semilinear  nested logit model.
We use both the Poisson estimator and the minimum-distance estimator
on the former model, and only the minimum-distance estimator on the latter.
"""

from typing import List, Tuple
import matplotlib.pyplot as plt
import numpy as np
from numpy.random import SeedSequence
import pandas as pd
import seaborn as sns

from .utils import nprepeat_col, nprepeat_row, print_stars
from .choo_siow import entropy_choo_siow
from .entropy import EntropyFunctions
from .min_distance_utils import MDEResults
from .min_distance import estimate_semilinear_mde
from .model_classes import ChooSiowPrimitives, NestedLogitPrimitives
from .nested_logit import setup_standard_nested_logit
from .poisson_glm import PoissonGLMResults, choo_siow_poisson_glm


def choo_siow_simul(
    phi_bases: np.ndarray,
    n: np.ndarray,
    m: np.ndarray,
    true_coeffs: np.ndarray,
    n_households: int,
    n_simuls: int,
    seed: int = None,
) -> Tuple[List[MDEResults], List[PoissonGLMResults]]:
    """
    Monte Carlo simulation of the minimum distance and Poisson estimators
    for the Choo and Siow model

    Args:
        phi_bases: an (X,Y,K) array of bases
        n: an X-vector, margins for men
        m: an Y-vector, margins for women
        true_coeffs: a K-vector, the true values of
            the coefficients of the bases
        n_households: the number of households
        n_simuls: the number of samples for the simulation
        seed: an integer seed for the random number generator

    Returns:
        the lists of results for the min distance estimator
        and the Poisson GLM estimator

    """

    Phi = phi_bases @ true_coeffs
    choo_siow_instance = ChooSiowPrimitives(Phi, n, m)

    ss = SeedSequence(seed)
    child_seeds = ss.spawn(n_simuls)

    min_distance_results = []
    poisson_results = []
    for s in range(n_simuls):
        mus_sim = choo_siow_instance.simulate(n_households, child_seeds[s])
        mde_results = estimate_semilinear_mde(mus_sim, phi_bases, entropy_choo_siow)
        min_distance_results.append(mde_results)
        poisson_glm_results = choo_siow_poisson_glm(mus_sim, phi_bases, verbose=1)
        poisson_results.append(poisson_glm_results)
        print(f"\nChoo-Siow: estimates for sample {s}:")
        print("    MDE:")
        print(mde_results.estimated_coefficients)
        print("    Poisson:")
        print(poisson_glm_results.estimated_beta)
    return min_distance_results, poisson_results


def nested_logit_simul(
    phi_bases: np.ndarray,
    n: np.ndarray,
    m: np.ndarray,
    entropy_model: EntropyFunctions,
    true_alphas: np.ndarray,
    true_betas: np.ndarray,
    n_households: int,
    n_simuls: int,
    seed: int = None,
) -> List[MDEResults]:
    """
    Monte Carlo simulation of the minimum distance estimator
    for the nested logit

    Args:
        phi_bases: an (X,Y,K) array of bases
        n: an X-vector, margins for men
        m: an Y-vector, margins for women
        entropy_model: the nested logit specification
        true_alphas: an (n_rhos+n_deltas)-vector,
            the true values of the nests parameters
        true_betas: a K-vector,
            the true values of the coefficients of the bases
        n_households: the number of households
        n_simuls: the number of samples for the simulation
        seed: an integer seed for the random number generator

    Returns:
        the list of results for the min distance estimator
    """

    Phi = phi_bases @ true_betas
    nests_for_each_x, nests_for_each_y = entropy_model.more_params
    nested_logit_instance = NestedLogitPrimitives(
        Phi, n, m, nests_for_each_x, nests_for_each_y, true_alphas
    )

    ss = SeedSequence(seed)
    child_seeds = ss.spawn(n_simuls)

    min_distance_results = []
    for s in range(n_simuls):
        mus_sim = nested_logit_instance.simulate(n_households, child_seeds[s])
        mde_result = estimate_semilinear_mde(
            mus_sim,
            phi_bases,
            entropy_model,
            more_params=entropy_model.more_params,
        )
        min_distance_results.append(mde_result)
        print(f"\nNested logit: MDE estimates for sample {s}:")
        print(mde_result.estimated_coefficients)
    return min_distance_results


if __name__ == "__main__":

    """we draw n_simuls samples of n_households households"""
    n_households = 100_000
    n_simuls = 1000

    run_choo_siow = True
    run_nested_logit = False

    plot_choo_siow = True
    plot_nested_logit = False

    # integer to select a variant; None to do the central scenario
    do_variant = None

    X, Y, K = 20, 20, 8
    # set of 8 basis functions
    phi_bases = np.zeros((X, Y, K))
    phi_bases[:, :, 0] = 1.0
    vec_x = np.arange(X)
    vec_y = np.arange(Y)
    phi_bases[:, :, 1] = nprepeat_col(vec_x, Y)
    phi_bases[:, :, 2] = nprepeat_row(vec_y, X)
    phi_bases[:, :, 3] = phi_bases[:, :, 1] * phi_bases[:, :, 1]
    phi_bases[:, :, 4] = phi_bases[:, :, 1] * phi_bases[:, :, 2]
    phi_bases[:, :, 5] = phi_bases[:, :, 2] * phi_bases[:, :, 2]
    for i in range(X):
        for j in range(i, Y):
            phi_bases[i, j, 6] = 1
            phi_bases[i, j, 7] = i - j
    true_betas = np.array([1.0, 0.0, 0.0, -0.01, 0.02, -0.01, 0.5, 0.0])
    str_variant = ""

    if do_variant is not None:
        if do_variant == 1:
            X, Y, K = 10, 10, 4
            phi_bases = phi_bases[:X, :Y, :K]
            true_betas = true_betas[:K]
        elif do_variant == 2:
            X, Y, K = 4, 5, 6
            phi_bases = phi_bases[:X, :Y, :K]
            true_betas = true_betas[:K]
        elif do_variant == 3:
            X, Y, K = 20, 20, 2
            phi_bases = phi_bases[:X, :Y, :K]
            true_betas = true_betas[:K]
        str_variant = f"_v{do_variant}"

    t = 0.2
    n = np.logspace(start=0, base=1 - t, stop=X - 1, num=X)
    m = np.logspace(start=0, base=1 - t, stop=Y - 1, num=Y)

    beta_names = [f"beta[{i}]" for i in range(1, K + 1)]

    seed = 5456456

    if run_choo_siow:
        choo_siow_results_file = (
            "choo_siow_simul_results"
            + f"{str_variant}_N{n_households}_seed{seed}"
            + ".csv"
        )
        min_distance_results, poisson_results = choo_siow_simul(
            phi_bases, n, m, true_betas, n_households, n_simuls, seed
        )
        mde_estimated_beta = np.zeros((n_simuls, K))
        poisson_estimated_beta = np.zeros((n_simuls, K))
        mde_stderrs_beta = np.zeros((n_simuls, K))
        poisson_stderrs_beta = np.zeros((n_simuls, K))
        K2 = 2 * K
        n_rows = K2 * n_simuls
        simulation = np.zeros(n_rows, dtype=int)
        estimate = np.zeros(n_rows)
        stderrs = np.zeros(n_rows)
        estimator = []
        coefficient = []
        beg_s = 0
        for s in range(n_simuls):
            mde_resus_s = min_distance_results[s]
            poisson_resus_s = poisson_results[s]
            mde_estimated_beta[s, :] = mde_resus_s.estimated_coefficients
            poisson_estimated_beta[s, :] = poisson_resus_s.estimated_beta
            mde_stderrs_beta[s, :] = mde_resus_s.stderrs_coefficients
            poisson_stderrs_beta[s, :] = poisson_resus_s.stderrs_beta
            slice_K2 = slice(beg_s, beg_s + K2)
            simulation[slice_K2] = s
            estimator.extend(["Minimum distance"] * K)
            estimator.extend(["Poisson"] * K)
            coefficient.extend(beta_names)
            coefficient.extend(beta_names)
            slice_K = slice(beg_s, beg_s + K)
            estimate[slice_K] = mde_estimated_beta[s, :]
            stderrs[slice_K] = mde_stderrs_beta[s, :]
            slice_K_K2 = slice(beg_s + K, beg_s + K2)
            estimate[slice_K_K2] = poisson_estimated_beta[s, :]
            stderrs[slice_K_K2] = poisson_stderrs_beta[s, :]
            beg_s += K2
        true_values = np.tile(true_betas, 2 * n_simuls)

        choo_siow_results = pd.DataFrame(
            {
                "Simulation": simulation,
                "Estimator": estimator,
                "Parameter": coefficient,
                "Estimate": estimate,
                "Standard Error": stderrs,
                "True value": true_values,
            }
        )

        choo_siow_results.to_csv(choo_siow_results_file)

    if plot_choo_siow:
        choo_siow_results_file = (
            "choo_siow_simul_results"
            + f"{str_variant}_N{n_households}_seed{seed}"
            + ".csv"
        )
        choo_siow_results = pd.read_csv(choo_siow_results_file)

        # discard outliers
        beta_err = np.array([2.0, 0.5, 0.5, 0.1, 0.1, 0.1, 2.0, 0.5])[:K]
        beta_min = true_betas - beta_err
        beta_max = true_betas + beta_err
        beta_min2 = np.tile(beta_min, 2)
        beta_max2 = np.tile(beta_max, 2)
        beta_min_large = np.tile(beta_min, 2 * n_simuls)
        beta_max_large = np.tile(beta_max, 2 * n_simuls)

        beta_estimates = choo_siow_results.Estimate.values
        outlier_mask = (beta_estimates < beta_min_large) | (
            beta_estimates > beta_max_large
        )
        outlier_simuls = choo_siow_results["Simulation"][outlier_mask].unique()

        print_stars(
            f"We have a total of {len(outlier_simuls)} outliers"
            + f" out of {n_simuls} simulations."
        )
        # for s in outlier_simuls:
        #     print(f" Outlier for simulation {s}:")
        #     mask_s = choo_siow_results["Simulation"] == s
        #     outlier_s = choo_siow_results[mask_s]
        #     estimates_s = outlier_s.Estimate.values
        #     estimators_s = outlier_s.Estimator.values
        #     parameters_s = outlier_s.Parameter.values
        #     wrong_s = (estimates_s > beta_max2) | (estimates_s < beta_min2)
        #     for i, w in enumerate(wrong_s):
        #         if w:
        #             print(
        #                 f"{estimators_s[i]} failed for {parameters_s[i]}"
        #                 + f" with value {estimates_s[i]}"
        #             )

        choo_siow_cleaned_results = choo_siow_results[
            ~choo_siow_results.Simulation.isin(outlier_simuls)
        ]

        g = sns.FacetGrid(
            data=choo_siow_cleaned_results,
            sharex=False,
            sharey=False,
            hue="Estimator",
            col="Parameter",
            col_wrap=2,
        )
        g.map(sns.kdeplot, "Estimate")
        g.set_titles("{col_name}")
        for true_val, ax in zip(true_betas, g.axes.ravel()):
            ax.vlines(true_val, *ax.get_ylim(), color="k", linestyles="dashed")
        g.add_legend()

        plt.savefig(
            "choo_siow_simul_results"
            + f"{str_variant}_N{n_households}_seed{seed}"
            + ".png"
        )

        gs = sns.FacetGrid(
            data=choo_siow_cleaned_results,
            sharex=False,
            sharey=False,
            hue="Estimator",
            col="Parameter",
            col_wrap=2,
        )
        gs.map(sns.kdeplot, "Standard Error")
        gs.set_titles("{col_name}")
        gs.add_legend()

        plt.savefig(
            "choo_siow_simul_results_stderrs"
            + f"{str_variant}_N{n_households}_seed{seed}"
            + ".png"
        )

    if run_nested_logit:
        # Nests and nest parameters for our two-level nested logit
        #  0 is the first nest;
        #   all other nests and nest parameters are type-independent
        # each x has the same nests over 1, ..., Y
        nests_for_each_y = [
            list(range(1, Y // 2 + 1)),
            list(range(Y // 2 + 1, Y + 1)),
        ]
        # each y has the same nests over 1, ..., X
        nests_for_each_x = [
            list(range(1, X // 2 + 1)),
            list(range(X // 2 + 1, X + 1)),
        ]

        entropy_nested_logit, _ = setup_standard_nested_logit(
            nests_for_each_x, nests_for_each_y
        )

        n_rhos, n_deltas = len(nests_for_each_x), len(nests_for_each_y)
        n_alphas = n_rhos + n_deltas

        true_alphas = np.full(n_alphas, 0.5)

        nested_logit_results_file = (
            f"nested_logit_simul_results_N{n_households}_seed{seed}.csv"
        )

        min_distance_results = nested_logit_simul(
            phi_bases,
            n,
            m,
            entropy_nested_logit,
            true_alphas,
            true_betas,
            n_households,
            n_simuls,
        )

        true_coeffs = np.concatenate((true_alphas, true_betas))
        n_pars = n_alphas + K

        alpha_names = [f"rho[{i+1}]" for i in range(n_rhos)] + [
            f"delta[{i+1}]" for i in range(n_deltas)
        ]
        alphabeta_names = alpha_names + beta_names

        mde_estimated_coeffs = np.zeros((n_simuls, n_pars))
        mde_stderrs = np.zeros((n_simuls, n_pars))
        n_rows = n_pars * n_simuls
        simulation = np.zeros(n_rows, dtype=int)
        estimate = np.zeros(n_rows)
        stderrs = np.zeros(n_rows)
        coefficient = []
        beg_s = 0
        for s in range(n_simuls):
            resus_s = min_distance_results[s]
            mde_estimated_coeffs[s, :] = resus_s.estimated_coefficients
            mde_stderrs[s, :] = resus_s.stderrs_coefficients
            slice_s = slice(beg_s, beg_s + n_pars)
            simulation[slice_s] = s
            coefficient.extend(alphabeta_names)
            estimate[slice_s] = mde_estimated_coeffs[s, :]
            stderrs[slice_s] = mde_stderrs[s, :]
            beg_s += n_pars

        true_values = np.tile(true_coeffs, n_simuls)

        nested_logit_results = pd.DataFrame(
            {
                "Simulation": simulation,
                "Parameter": coefficient,
                "Estimate": estimate,
                "Standard Error": stderrs,
                "True value": true_values,
            }
        )

        nested_logit_results.to_csv(nested_logit_results_file)

    if plot_nested_logit:
        nested_logit_results_file = (
            f"nested_logit_simul_results_N{n_households}_seed{seed}.csv"
        )

        nested_logit_results = pd.read_csv(nested_logit_results_file)

        g = sns.FacetGrid(
            data=nested_logit_results,
            sharex=False,
            sharey=False,
            col="Parameter",
            col_wrap=2,
        )
        g.map(sns.kdeplot, "Estimate")
        g.set_titles("{col_name}")
        for true_val, ax in zip(true_values, g.axes.ravel()):
            ax.vlines(true_val, *ax.get_ylim(), color="k", linestyles="dashed")
        g.add_legend()

        plt.savefig(f"nested_logit_results_N{n_households}_seed{seed}.png")

        gs = sns.FacetGrid(
            data=nested_logit_results,
            sharex=False,
            sharey=False,
            col="Parameter",
            col_wrap=2,
        )
        gs.map(sns.kdeplot, "Standard Error")
        gs.set_titles("{col_name}")
        gs.add_legend()

        plt.savefig(f"nested_logit_results_stderrs_N{n_households}_seed{seed}.png")



"""This module contains some utility programs used by the package."""

import sys
from collections import namedtuple
from math import exp, log, sqrt
from traceback import extract_stack
from typing import Callable, List, Optional, Tuple, Union
import numpy as np
import scipy.stats as sts

ScalarFunctionAndGradient = Callable[
    [np.ndarray, List, Optional[bool]], Union[float, Tuple[float, np.ndarray]]
]
"""Type of f(v, args, gr) that returns a scalar value
and also a gradient if gr is True"""

# for central numerical derivatives
_EPS = 1e-4
_TWO_EPS = 2.0 * _EPS


def bs_name_func(back: int = 2) -> str:
    """Get the name of the current function, or further back in the stack

    Args:
        back: 2 for the current function, 3 for the function that called it, etc

    Returns:
        the name of the function requested
    """
    stack = extract_stack()
    func_name = stack[-back][2]
    return func_name


def print_stars(title: str = None, n: int = 70) -> None:
    """Prints a starred line, or two around the title

    Args:
        title:  an optional title
        n: the number of stars on the line

    Returns:
        nothing
    """
    line_stars = "*" * n
    print()
    print(line_stars)
    if title:
        print(title.center(n))
        print(line_stars)
    print()


def bs_error_abort(msg: str = "error, aborting") -> None:
    """Report error and exits with code 1

    Args:
        msg: specifies the error message

    Returns:
        nothing
    """
    print_stars(f"{bs_name_func(3)}: {msg}")
    sys.exit(1)


def test_vector(x: np.ndarray, fun_name: str = None) -> int:
    """Tests that `x` is a vector; aborts otherwise

    Args:
        x: a potential vector
        fun_name: the name of the calling function

    Returns:
        the size of `x` if it is a vector
    """
    fun_str = ["" if fun_name is None else fun_name + ":"]
    if not isinstance(x, np.ndarray):
        bs_error_abort(f"{fun_str} x should be a Numpy array")
    ndims_x = x.ndim
    if ndims_x != 1:
        bs_error_abort(f"{fun_str} x should have one dimension, not {ndims_x}")
    return x.size


def test_matrix(x: np.ndarray, fun_name: str = None) -> Tuple[int, int]:
    """Tests that `x` is a matrix; aborts otherwise

    Args:
        x: a potential matrix
        fun_name: the name of the calling function

    Returns:
        the shape of `x` if it is a matrix
    """
    fun_str = ["" if fun_name is None else fun_name + ":"]
    if not isinstance(x, np.ndarray):
        bs_error_abort(f"{fun_str} x should be a Numpy array")
    ndims_x = x.ndim
    if ndims_x != 2:
        bs_error_abort(f"{fun_str} x should have two dimensions, not {ndims_x}")
    return x.shape


def describe_array(v: np.ndarray, name: str = "The array") -> namedtuple:
    """Descriptive statistics on an array interpreted as a vector

    Args:
        v: the array
        name: its name

    Returns:
        a `DescribeResult` namedtuple
    """
    print_stars(f"{name} has:")
    d = sts.describe(v, None)
    print(f"Number of elements: {d.nobs}")
    print(f"Minimum: {d.minmax[0]}")
    print(f"Maximum: {d.minmax[1]}")
    print(f"Mean: {d.mean}")
    print(f"Stderr: {sqrt(d.variance)}")
    return d


def nprepeat_col(v: np.ndarray, n: int) -> np.ndarray:
    """Creates a matrix with `n` columns, all equal to `v`

    Args:
        v: a vector of size `m`
        n: the number of columns requested

    :return: a matrix of shape `(m, n)`
    """
    _ = test_vector(v, "nprepeat_col")
    return np.repeat(v[:, np.newaxis], n, axis=1)


def nprepeat_row(v: np.ndarray, m: int) -> np.ndarray:
    """
    Creates a matrix with `m` rows, all equal to `v`

    Args:
        v: a vector of size `n`
        m: the number of rows requested

    Returns:
        a matrix of shape `(m, n)`
    """
    _ = test_vector(v, "nprepeat_row")
    return np.repeat(v[np.newaxis, :], m, axis=0)


def npmaxabs(a: np.ndarray) -> float:
    """The maximum absolute value in an array

    Args:
        a: the array

    Returns:
        $\max{\vert a \vert}$
    """
    return np.max(np.abs(a))


def nplog(
    a: np.ndarray, deriv: bool = False, eps: float = 1e-30, verbose: bool = False
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """$C^2$ extension of  $\ln(a)$ below `eps`

    Args:
        a: a Numpy array
        deriv: if `True`, the first derivative is also returned
        eps: a lower bound
        verbose: whether diagnoses are printed

    Returns:
        $\ln(a)$ $C^2$-extended below `eps`,
        with its derivative if `deriv` is `True`
    """
    if np.min(a) > eps:
        loga = np.log(a)
        return [loga, 1.0 / a] if deriv else loga
    else:
        logarreps = np.log(np.maximum(a, eps))
        logarr_smaller = log(eps) - (eps - a) * (3.0 * eps - a) / (2.0 * eps * eps)
        if verbose:
            n_small_args = np.sum(a < eps)
            if n_small_args > 0:
                finals = "s" if n_small_args > 1 else ""
                print(
                    f"nplog: {n_small_args} argument{finals} smaller than {eps}: mini = {np.min(a)}"
                )
        loga = np.where(a > eps, logarreps, logarr_smaller)
        if deriv:
            der_logarreps = 1.0 / np.maximum(a, eps)
            der_logarr_smaller = (2.0 * eps - a) / (eps * eps)
            der_loga = np.where(a > eps, der_logarreps, der_logarr_smaller)
            return loga, der_loga
        else:
            return loga


def npexp(
    a: np.ndarray, deriv: bool = False, bigx: float = 30.0, verbose: bool = False
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """
    $C^2$ extension of  $\exp(a)$ above `bigx`

    Args:
        a: a Numpy array
        deriv: if `True`, the first derivative is also returned
        bigx: an upper bound
        verbose: whether diagnoses are printed

    Returns:
        bigx: upper bound $\exp(a)$  $C^2$-extended above `bigx`,
        with its derivative if `deriv` is `True`
    """
    if np.max(a) < bigx:
        expa = np.exp(a)
        return [expa, expa] if deriv else expa
    else:
        exparr = np.exp(np.minimum(a, bigx))
        ebigx = exp(bigx)
        darr = a - bigx
        exparr_larger = ebigx * (1.0 + darr * (1.0 + 0.5 * darr))
        if verbose:
            n_large_args = np.sum(a > bigx)
            if n_large_args > 0:
                finals = "s" if n_large_args > 1 else ""
                print(
                    f"npexp: {n_large_args} argument{finals} larger than {bigx}: maxi = {np.max(a)}"
                )
        expa = np.where(a < bigx, exparr, exparr_larger)
        if deriv:
            der_exparr = np.exp(np.minimum(a, bigx))
            der_exparr_larger = ebigx * (1.0 + darr)
            der_expa = np.where(a < bigx, der_exparr, der_exparr_larger)
            return expa, der_expa
        else:
            return expa


def nppow(
    a: np.ndarray, b: Union[int, float, np.ndarray], deriv: bool = False
) -> Union[np.array, Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """Evaluates $a^b$ element-by-element

    Args:
        a: a Numpy array
        b: if an array, it should have the same shape as `a`
        deriv: if `True`, the first derivatives wrt `a` and `b`
            are also returned

    Returns:
        an array of the same shape as `a`, and if `deriv` is `True`,
        the derivatives wrt `a` and `b`
    """
    mina = np.min(a)
    if mina <= 0:
        bs_error_abort("All elements of a must be positive!")

    if isinstance(b, (int, float)):
        a_pow_b = a ** b
        if deriv:
            return (a ** b, b * a_pow_b / a, a_pow_b * log(a))
        else:
            return a_pow_b
    else:
        if a.shape != b.shape:
            bs_error_abort(f"a has shape {a.shape} and b has shape {b.shape}")
        avec = a.ravel()
        bvec = b.ravel()
        a_pow_b = avec ** bvec
        if deriv:
            der_wrt_a = a_pow_b * bvec / avec
            der_wrt_b = a_pow_b * nplog(avec)
            return (
                a_pow_b.reshape(a.shape),
                der_wrt_a.reshape(a.shape),
                der_wrt_b.reshape(a.shape),
            )
        else:
            return a_pow_b.reshape(a.shape)


def der_nppow(a: np.array, b: Union[int, float, np.array]) -> np.array:
    """
    evaluates the derivatives in a and b of element-by-element $a^b$

    :param np.array a:

    :param Union[int, float, np.array] b: if an array,
       should have the same shape as `a`

    :return: a pair of two arrays of the same shape as `a`
    """

    mina = np.min(a)
    if mina <= 0:
        print_stars("All elements of a must be positive in der_nppow!")
        sys.exit(1)

    if isinstance(b, (int, float)):
        a_pow_b = a ** b
        return (b * a_pow_b / a, a_pow_b * log(a))
    else:
        if a.shape != b.shape:
            print_stars("nppow: b is not a number or an array of the same shape as a!")
            sys.exit(1)
        avec = a.ravel()
        bvec = b.ravel()
        a_pow_b = avec ** bvec
        der_wrt_a = a_pow_b * bvec / avec
        der_wrt_b = a_pow_b * nplog(avec)
        return (der_wrt_a.reshape(a.shape), der_wrt_b.reshape(a.shape))


def check_gradient_scalar_function(
    fg: ScalarFunctionAndGradient,
    p: np.ndarray,
    args: List,
    mode: str = "central",
    EPS: float = 1e-6,
) -> Tuple[np.ndarray, np.ndarray]:
    """Checks the gradient of a scalar function.

    Args:
        fg: should return the scalar value, and the gradient if its `gr` argument is `True`
        p: where we are checking the gradient
        args: other arguments passed to `fg`
        mode: "central" or "forward" derivatives
        EPS: the step for forward or central derivatives

    Returns:
        the analytic and numeric gradients
    """
    f0, f_grad = fg(p, args, gr=True)

    print_stars("checking the gradient: analytic, numeric")

    g = np.zeros_like(p)
    if mode == "central":
        for i, x in enumerate(p):
            p1 = p.copy()
            p1[i] = x + EPS
            f_plus = fg(p1, args, gr=False)
            p1[i] -= 2.0 * EPS
            f_minus = fg(p1, args, gr=False)
            g[i] = (f_plus - f_minus) / (2.0 * EPS)
            print(f"{i}: {f_grad[i]}, {g[i]}")
    elif mode == "forward":
        for i, x in enumerate(p):
            p1 = p.copy()
            p1[i] = x + EPS
            f_plus = fg(p1, args, gr=False)
            g[i] = (f_plus - f0) / EPS
            print(f"{i}: {f_grad[i]}, {g[i]}")
    else:
        bs_error_abort("mode must be 'central' or 'forward'")

    return f_grad, g


if __name__ == "__main__":
    """run some tests"""

    eps = 1e-6

    arr = np.array([-0.001, 0.0, 1e-30, 0.001])
    print(f"x = {arr}")
    print_stars("numpy extended log(x)")
    loga, der_loga = nplog(arr, deriv=True)
    print(f"   value = {loga}")
    print(f"   derivative = {der_loga}")

    arr = np.array(np.arange(6)).reshape((2, 3))
    describe_array(arr, "arr")

    args_exp = np.array([[10.0, 30.0], [32.0, -1.0]])
    print(f"\n\nextended exponential of {args_exp}:")
    expa, der_expa = npexp(args_exp, deriv=True)
    print(expa)
    print("\nits first derivative should be:")
    print((npexp(args_exp + eps) - npexp(args_exp - eps)) / (2.0 * eps))
    print("       it is:")
    print(der_expa)

    print_stars("Testing rows and columns repeats")
    v = np.arange(3)
    vm = nprepeat_row(v, 2)
    vn = nprepeat_col(v, 4)
    print("v=:")
    print(v)
    print("2 rows of v:")
    print(vm)
    print("4 columns of v:")
    print(vn)

    def fg(p, args, gr=False):
        a = args[0]
        p0, p1 = p[0], p[1]
        f = p0 ** 2 + p1 ** a
        if gr:
            g = np.zeros(2)
            g[0] = 2.0 * p0
            g[1] = a * (p1 ** (a - 1))
            return f, g
        else:
            return f

    check_gradient_scalar_function(fg, np.array([2.0, 4.0]), [3.0])



from dataclasses import dataclass
from typing import List, Optional, Tuple
import numpy as np

from .utils import (
    bs_error_abort,
    npexp,
    npmaxabs,
    nppow,
    print_stars,
    test_matrix,
    test_vector,
)
from .matching_utils import (
    Matching,
    _change_indices,
    _compute_margins,
    _find_nest_of,
    _simulate_sample_from_mus,
)
from .ipfp_solvers import ipfp_homoskedastic_solver


@dataclass
class ChooSiowPrimitives:
    Phi: np.ndarray
    n: np.ndarray
    m: np.ndarray
    mus: Optional[Matching] = None

    def __post_init__(self):
        X, Y = test_matrix(self.Phi)
        Xn = test_vector(self.n)
        Ym = test_vector(self.m)
        if Xn != X:
            bs_error_abort(f"Phi is a ({X}, {Y}) matrix but n has {Xn} elements.")
        if Ym != Y:
            bs_error_abort(f"Phi is a ({X}, {Y}) matrix but m has {Ym} elements.")

    def ipfp_solve(self) -> Matching:
        mus, err_x, err_y = ipfp_homoskedastic_solver(self.Phi, self.n, self.m)
        muxy, mux0, mu0y = mus
        n, m = _compute_margins(muxy, mux0, mu0y)
        return Matching(muxy, n, m)

    def simulate(self, n_households: int, seed: int = None) -> Matching:
        self.n_households = n_households
        self.mus = self.ipfp_solve()
        mus_sim = _simulate_sample_from_mus(self.mus, n_households, seed)
        return mus_sim

    def describe(self):
        X, Y = self.Phi.shape
        print_stars("We are working with a Choo and Siow homoskedastic market")
        print(f"\t\t we have {int(X)} types of men and {int(Y)} types of women")
        print(f"\t\t and a total of {int(self.n_households)} households")


@dataclass
class NestedLogitPrimitives:
    Phi: np.ndarray
    n: np.ndarray
    m: np.ndarray
    nests_for_each_x: List[
        List[int]
    ]  # given by user, e.g. [[1, 3], [2,4]] has y=1 and y=3 in first nest
    nests_for_each_y: List[List[int]]
    nests_over_Y: List[
        List[int]
    ]  # rebased to zero: the above example becomes [[0, 2], [1,3]]
    nests_over_X: List[List[int]]
    i_nest_of_x: List[int]  # mapping x -> n'
    i_nest_of_y: List[int]  # mapping y -> n
    n_alphas: int
    mus: Optional[Matching] = None
    true_alphas: Optional[np.ndarray] = None

    def __init__(
        self,
        Phi: np.ndarray,
        n: np.ndarray,
        m: np.ndarray,
        nests_for_each_x: List[List[int]],
        nests_for_each_y: List[List[int]],
        true_alphas: Optional[np.ndarray] = None,
    ):
        """
        We only model two-level nested logit, with {0} as the first nest,
        and nests and nests parameters that do not depend on the type.

        Args:
            Phi: the (X,Y) joint surplus matrix
            n: the X-vector of men margins
            m: the X-vector of women margins
            nests_for_each_x: the composition of the nests over 1...Y, a list of r lists
            nests_for_each_y: the composition of the nests over 1...X, a list of d lists
            true_alphas: the true nest parameters, if any; should be an (r+d)-vector
        """
        X, Y = test_matrix(Phi)
        Xn = test_vector(n)
        Ym = test_vector(m)

        # we need to rebase the indices to zero
        self.nests_over_X = _change_indices(nests_for_each_y)
        self.nests_over_Y = _change_indices(nests_for_each_x)

        self.n_alphas = len(nests_for_each_y) + len(nests_for_each_x)

        if Xn != X:
            bs_error_abort(f"Phi is a ({X}, {Y}) matrix but n has {Xn} elements.")
        if Ym != Y:
            bs_error_abort(f"Phi is a ({X}, {Y}) matrix but m has {Ym} elements.")

        if true_alphas is not None:
            alpha_size = test_vector(true_alphas)
            if alpha_size != self.n_alphas:
                bs_error_abort(
                    f"true_alphas shoud have {self.n_alphas} elements, not {alpha_size}"
                )

        self.Phi = Phi
        self.n = n
        self.m = m
        self.true_alphas = true_alphas
        self.nests_for_each_x = nests_for_each_x
        self.nests_for_each_y = nests_for_each_y

        # check that every x is in a nest, and just once
        nests_check = []
        i_nest_of_x = np.zeros(X, int)
        for x in range(X):
            i_nest_of_x[x] = _find_nest_of(self.nests_over_X, x)
            nests_check.append(i_nest_of_x[x])
        if -1 in nests_check or len(set(nests_check)) != len(nests_for_each_y):
            bs_error_abort("Check your nests_for_each_y")
        # check that every y is in a nest, and just once
        nests_check = []
        i_nest_of_y = np.zeros(Y, int)
        for y in range(Y):
            i_nest_of_y[y] = _find_nest_of(self.nests_over_Y, y)
            nests_check.append(i_nest_of_y[y])
        if -1 in nests_check or len(set(nests_check)) != len(nests_for_each_x):
            bs_error_abort("Check your nests_for_each_x")

        self.i_nest_of_x = i_nest_of_x
        self.i_nest_of_y = i_nest_of_y

    def __str__(self):
        X, Y = self.Phi.shape
        nmen, nwomen = np.sum(self.n), np.sum(self.m)
        repr_str = (
            f"This is a 2-level nested logit with {nmen} men of {X} types"
            + f" and {nwomen} women of {Y} types.\n"
        )
        repr_str += (
            f" We have {self.n_nests_over_Y} nests over 1...Y "
            + f" and {self.n_nests_over_X} nests over 1...X,\n"
        )
        if self.true_alphas is None:
            repr_str += "     with unspecified nests parameters."
        else:
            alpha_vals = self.true_alphas
            repr_str += "     with respective nests parameters:\n"
            repr_str += f"    {alpha_vals[:self.n_nests_over_Y]}\n"
            repr_str += f" and {alpha_vals[self.n_nests_over_Y:]}\n"
        print_stars(repr_str)

    def ipfp_nested_logit_solver(
        self, tol: float = 1e-9, verbose: bool = False, maxiter: int = 1000
    ) -> Tuple[Matching, np.ndarray, np.ndarray]:
        """Solves for equilibrium in a two-level nested logit market
        given systematic surplus and margins and nests parameters;
        does not compute the gradient of the matching patterns

        Args:
            tol: tolerance on change in solution
            verbose: if `True`, prints information
            maxiter: maximum number of iterations

        Returns:
             (muxy, mux0, mu0y): the matching patterns
             marg_err_x, marg_err_y: the errors on the margins
        """
        alphas = self.true_alphas
        if alphas is None:
            bs_error_abort("cannot solve without nest parameters")

        n_rhos = len(self.nests_over_Y)
        n_deltas = len(self.nests_over_X)
        rhos = alphas[:n_rhos]
        deltas = alphas[n_rhos:]

        #############################################################################
        # we solve the equilibrium equations
        #   starting with a reasonable initial point  muxy, mux0, mu0y = bigc
        #   it is important that it fit the number of individuals
        #############################################################################

        n, m = self.n, self.m
        X, Y = n.size, m.size

        nests_over_X, nests_over_Y = self.nests_over_X, self.nests_over_Y
        i_nest_of_x, i_nest_of_y = self.i_nest_of_x, self.i_nest_of_y

        rho_vals = rhos[i_nest_of_y]  # rho(n) for y in n in the paper
        delta_vals = deltas[i_nest_of_x]  # delta(n') for x in n' in the paper

        ephi = npexp(self.Phi / np.add.outer(delta_vals, rho_vals))

        # initial values
        nindivs = np.sum(n) + np.sum(m)
        bigc = nindivs / (X + Y + 2.0 * np.sum(ephi))

        mux0, mu0y, muxy = (
            np.full(X, bigc),
            np.full(Y, bigc),
            np.full((X, Y), bigc),
        )
        muxn = np.zeros((X, n_rhos))
        for i_nest_y, nest_y in enumerate(nests_over_Y):
            muxn[:, i_nest_y] = np.sum(muxy[:, nest_y], 1)
        muny = np.zeros((n_deltas, Y))
        for i_nest_x, nest_x in enumerate(nests_over_X):
            muny[i_nest_x, :] = np.sum(muxy[nest_x, :], 0)

        err_diff = bigc
        tol_diff = tol * bigc
        tol_newton = tol
        max_newton = 2000
        MIN_REST = 1e-4 * bigc  # used to bound mus below in the Newton iterations

        niter = 0
        while (err_diff > tol_diff) and (niter < maxiter):  # IPFP main loop
            # Newton iterates for men
            err_newton = bigc
            i_newton = 0
            while err_newton > tol_newton:
                gbar = np.zeros(
                    (X, n_rhos)
                )  # this will be the $\bar{G}^x_n$ of the note
                gbar_pow = np.zeros((X, n_rhos))
                biga = np.zeros(X)  # this will be the $A_x$ of the note
                for i_nest_x, nest_x in enumerate(nests_over_X):
                    # i_nest_x is n' in the paper
                    delta_x = deltas[i_nest_x]
                    muny_x = muny[i_nest_x, :]  # mu(n', :)
                    for x in nest_x:
                        ephi_x = ephi[x, :]
                        for i_nest_y, nest_y in enumerate(nests_over_Y):
                            # i_nest_y is n in the paper
                            mu_n = muny_x[nest_y]
                            mu0_n = mu0y[nest_y]
                            evec_n = ephi_x[nest_y]
                            rho_n = rhos[i_nest_y]
                            sum_rd = rho_n + delta_x
                            mun_term = nppow(mu_n, (delta_x - 1.0) / sum_rd)
                            mu0_term = nppow(mu0_n, 1.0 / sum_rd)
                            gbar[x, i_nest_y] = np.sum(mun_term * mu0_term * evec_n)
                            gbar_pow[x, i_nest_y] = nppow(
                                gbar[x, i_nest_y], sum_rd / (delta_x + 1.0)
                            )
                            biga[x] += gbar_pow[x, i_nest_y]

                # now we take one Newton step for all types of men
                delta_vals1 = 1.0 + delta_vals
                mux0_term = nppow(mux0, 1.0 / delta_vals1)
                bigb = mux0_term * biga  # this is the $B_x$ of the note
                numer = n * delta_vals1 - delta_vals * bigb
                lower_bound = np.full(X, MIN_REST)
                mux0_new = mux0 * np.maximum(
                    numer / (delta_vals1 * mux0 + bigb), lower_bound
                )
                muxn_new = gbar_pow * mux0_term.reshape((-1, 1))

                mux0 = mux0_new
                muxn = muxn_new
                errxi = mux0 + np.sum(muxn, 1) - n
                err_newton = npmaxabs(errxi)
                i_newton += 1
                if i_newton > max_newton:
                    bs_error_abort(
                        f"Newton solver failed for men after {max_newton} iterations"
                    )

            if verbose:
                print(
                    f"Newton error on men is {err_newton} after {i_newton} iterations"
                )

            # Newton iterates for women
            err_newton = bigc
            i_newton = 0
            while err_newton > tol_newton:
                gbar = np.zeros((Y, n_deltas))
                gbar_pow = np.zeros((Y, n_deltas))
                biga = np.zeros(Y)
                for i_nest_y, nest_y in enumerate(nests_over_Y):
                    # i_nest_y is n in the paper
                    rho_y = rhos[i_nest_y]
                    muxn_y = muxn[:, i_nest_y]  # mu(:, n)
                    for y in nest_y:
                        ephi_y = ephi[:, y]
                        for i_nest_x, nest_x in enumerate(nests_over_X):
                            mu_n = muxn_y[nest_x]
                            mu0_n = mux0[nest_x]
                            evec_n = ephi_y[nest_x]
                            delta_n = deltas[i_nest_x]
                            sum_rd = rho_y + delta_n
                            mun_term = nppow(mu_n, (rho_n - 1.0) / sum_rd)
                            mu0_term = nppow(mu0_n, 1.0 / sum_rd)
                            gbar[y, i_nest_x] = np.sum(mun_term * mu0_term * evec_n)
                            gbar_pow[y, i_nest_x] = nppow(
                                gbar[y, i_nest_x], sum_rd / (1.0 + rho_y)
                            )
                            biga[y] += gbar_pow[y, i_nest_x]

                # now we take one Newton step for all types of women
                rho_vals1 = 1.0 + rho_vals
                mu0y_term = nppow(mu0y, 1.0 / rho_vals1)
                bigb = mu0y_term * biga
                numer = m * rho_vals1 - rho_vals * bigb
                lower_bound = np.full(Y, MIN_REST)
                mu0y_new = mu0y * np.maximum(
                    numer / (rho_vals1 * mu0y + bigb), lower_bound
                )
                muny_new = gbar_pow.T * mu0y_term

                mu0y = mu0y_new
                muny = muny_new
                erryi = mu0y + np.sum(muny, 0) - m
                err_newton = npmaxabs(erryi)
                i_newton += 1
                if i_newton > max_newton:
                    bs_error_abort(
                        f"Newton solver failed for women after {max_newton} iterations"
                    )

            if verbose:
                print(
                    f"Newton error on women is {err_newton} after {i_newton} iterations"
                )

            muxy = np.zeros((X, Y))
            for x in range(X):
                i_nest_x = i_nest_of_x[x]  # n'
                ephi_x = ephi[x, :]
                mux0_x = mux0[x]
                muxn_x = muxn[x, :]
                delta_x = delta_vals[x]
                muny_x = muny[i_nest_x, :]
                for y in range(Y):
                    i_nest_y = i_nest_of_y[y]  # n
                    mu0y_y = mu0y[y]
                    rho_y = rho_vals[y]
                    muxn_xy = muxn_x[i_nest_y]
                    muny_xy = muny_x[y]
                    mu_term = (
                        mux0_x
                        * mu0y_y
                        * (muxn_xy ** (rho_y - 1.0))
                        * (muny_xy ** (delta_x - 1.0))
                    )
                    muxy[x, y] = ephi_x[y] * (mu_term ** (1.0 / (delta_x + rho_y)))

            n_sim, m_sim = _compute_margins(muxy, mux0, mu0y)
            marg_err_x, marg_err_y = npmaxabs(n_sim - n), npmaxabs(m_sim - m)

            if verbose:
                print(
                    f"Margin error on men is {marg_err_x} "
                    f" after {niter} IPFP iterations"
                )
                print(
                    f"Margin error on women is {marg_err_y} "
                    f" after {niter} IPFP iterations"
                )
            err_diff = marg_err_x + marg_err_y
            niter += 1

        n_sim, m_sim = _compute_margins(muxy, mux0, mu0y)
        marg_err_x, marg_err_y = npmaxabs(n_sim - n), npmaxabs(m_sim - m)

        print(f"Margin error on men is {marg_err_x} after {niter} IPFP iterations")
        print(f"Margin error on women is {marg_err_y} after {niter} IPFP iterations")

        return Matching(muxy, n, m), marg_err_x, marg_err_y

    def ipfp_solve(self) -> Matching:
        if self.true_alphas is None:
            bs_error_abort(
                "true_alphas must be specified to solve the nested logit by IPFP."
            )
        self.mus, err_x, err_y = self.ipfp_nested_logit_solver(verbose=False)
        return self.mus

    def simulate(self, n_households: int) -> Matching:
        self.mus = self.ipfp_solve()
        mus_sim = _simulate_sample_from_mus(self.mus, n_households)
        return mus_sim
